import smbl
import snakemake
import os

from ._program import *

DEEZ = os.path.join(smbl.bin_dir,"deez")


##########################################
##########################################


class Deez(Program):

	@classmethod
	def get_installation_files(cls):
		return [DEEZ]

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]

	@classmethod
	def install(cls):
		cls.git_clone("git://github.com/sfu-compbio/deez","deez")
		cls.run_make("deez")
		cls.install_file("deez/deez",DEEZ)



import smbl
import snakemake
import os

from ._program import *

CMAKE = get_bin_file_path("cmake")


##########################################
##########################################


class CMake(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				CMAKE,
			]

	@classmethod
	def install(cls):
		fn=cls.download_file("http://cmake.org/files/v3.3/cmake-3.3.2.tar.gz","cmake.tar.gz")
		dir=cls.extract_tar(fn,strip=1)
		cls.shell('cd "{dir}" && ./bootstrap --prefix="{install_dir}"'.format(dir=dir,install_dir=smbl.smbl_dir))
		cls.run_make("")
		print("install")
		cls.shell('cd "{dir}" && make install'.format(dir=dir))

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]



#
# TODO:
#  - fix linking error in CygWin
#

import smbl
import snakemake
import os

from ._program import *

BCFTOOLS = get_bin_file_path("bcftools")
VCFUTILS = get_bin_file_path("vcfutils.pl")


##########################################
##########################################


class BcfTools(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				BCFTOOLS,
				VCFUTILS,
			]

	@classmethod
	def install(cls):
		gitdir_bcftools=cls.git_clone("http://github.com/samtools/bcftools","bcftools")
		gitdir_htslib=cls.git_clone("http://github.com/samtools/htslib","htslib")
		cls.run_make("bcftools")
		cls.install_file("bcftools/bcftools",BCFTOOLS)
		cls.install_file("bcftools/vcfutils.pl",VCFUTILS)

	@classmethod
	def supported_platforms(cls):
		return ["osx","linux"]



import snakemake
import smbl
import collections
import random
import os
import abc
import shutil

__PLUGINS = set()
#__RULES= set()

def register_plugin(plugin):
	__PLUGINS.add(plugin)

def get_registered_plugins():
	return sorted(list(__PLUGINS),key=lambda x:x.get_plugin_name())

def get_bin_file_path(program):
	return os.path.join(smbl.bin_dir,program)

##########################################
##########################################

def correct_samtools_make(makefile_fn):
	makefile_backup_fn = makefile_fn+".backup"
	if not os.path.isfile(makefile_backup_fn):
		with open(makefile_fn, 'r') as makefile:
			content = makefile.read()
			if smbl.utils.is_linux():
				content = content.replace("-lcurses","-lncurses")
			elif smbl.utils.is_cygwin():
				content = content.replace("-D_FILE_OFFSET_BITS=64","-D_FILE_OFFSET_BITS=64 -Dexpl=exp -Dlogl=log")
		with open(makefile_fn, 'w') as makefile:
			makefile.write(content)

##########################################
##########################################

class ProgramWatcher(type):
	def __init__(cls,name,bases,clsdict):
		if len(cls.mro()) == 3:
			register_plugin(cls)
			cls.src_dir=os.path.join(smbl.src_dir,cls.__name__)
		super(ProgramWatcher, cls).__init__(name, bases, clsdict)


class Program(metaclass=ProgramWatcher):
	src_dir=""
	verbosity=False
	#list of classes of dependencies

	def __init__(self):
		pass

	@classmethod
	def get_plugin_name(cls):
		return cls.__name__

	@classmethod
	def depends_on(cls):
		return []

	@classmethod
	def status_message(cls,message):
		smbl.messages.message(
			message=message,
			program="SMBL",
			subprogram=cls.get_plugin_name(),
		)

	@classmethod
	def shell(cls,command):
		if cls.verbosity:
			smbl.utils.shell(command)
		else:
			smbl.utils.shell("({}) > /dev/null".format(command))
		
	@classmethod
	def set_verbosity(cls,verbosity):
		cls.verbosity=verbosity		

	@classmethod
	def install_all_steps(cls):
		cls.status_message("Installation started")
		cls.install_pre()
		cls.install()
		cls.install_post()
		cls.status_message("Installation completed")

	@classmethod
	def install_pre(cls):
		assert len(set(cls.supported_platforms()))>0, ""
		assert set(cls.supported_platforms()).issubset(smbl.prog.all_platforms)
		if not cls.is_platform_supported():
			smbl.messages.error("Operating system '{}' is not supported".format(smbl.utils.get_platform()),program="SMBL")
			raise NotImplementedError("Unsupported OS")

		try:
			shutil.rmtree(cls.src_dir)
		except FileNotFoundError:
			pass
		cls.shell('mkdir -p "{src_dir}"'.format(src_dir=cls.src_dir))

	@classmethod
	def install(cls):
		raise NotImplementedError("Plugin error: install method must be defined")

	@classmethod
	def install_post(cls):
		try:
			shutil.rmtree(cls.src_dir)
		except FileNotFoundError:
			pass

	@classmethod
	def supported_platforms(cls):
		raise NotImplementedError("Plugin error: Supported platforms must be specified")

	@classmethod
	def is_platform_supported(cls):
		platform=smbl.utils.get_platform()
		if platform in cls.supported_platforms():
			return True
		else:
			return False

	@classmethod
	def get_installation_files(cls):
		raise NotImplementedError("Plugin error: Installation files must be specified")

	@classmethod
	def git_clone(cls,repository,dirname_short):
		cls.status_message("Cloning a GIT repository: "+repository)
		dirname_full=cls.abs_from_short(dirname_short)
		cls.shell('mkdir -p "{dir}"'.format(dir=dirname_full))
		cls.shell('git clone --recursive --depth 1 "{rep}" "{dir}"'.format(rep=repository,dir=dirname_full))
		return dirname_full

	@classmethod
	def svn_checkout(cls,repository,dirname_short):
		cls.status_message("Cloning a SVN repository: "+repository)
		dirname_full=cls.abs_from_short(dirname_short)
		cls.shell('mkdir -p "{dir}"'.format(dir=dirname_full))
		cls.shell('git svn clone "{rep}" "{dir}"'.format(rep=repository,dir=dirname_full))
		return dirname_full

	@classmethod
	def download_file(cls,address,filename_short):
		cls.status_message("Downloading a file: "+address)
		filename_full=cls.abs_from_short(filename_short)
		cls.shell('mkdir -p "{dir}"'.format(dir=os.path.dirname(filename_full)))
		try:
			cls.shell('curl -L --insecure -o "{fn}" "{address}"'.format(fn=filename_full,address=address))
		except:
			try:
				cls.shell('curl -L --insecure -o "{fn}" "{address}"'.format(fn=filename_full,address=address))
			except:
				try:
					cls.shell('wget --no-check-certificate --output-document="{fn}" "{address}"'.format(fn=filename_full,address=address))
				except:
					cls.shell('wget --no-check-certificate --output-document="{fn}" "{address}"'.format(fn=filename_full,address=address))

		return filename_full

	@classmethod
	def install_file(cls,filename_short,dest,executable=True):
		cls.status_message("Copying: "+cls.abs_from_short(filename_short))
		filename_full=cls.abs_from_short(filename_short)
		assert os.path.isfile(filename_full), "File which should be installed has not been created"
		cls.shell('cp "{source}" "{dest}"'.format(source=filename_full,dest=dest))
		if executable:
			cls.shell('chmod +x "{}"'.format(dest))

	@classmethod
	def extract_tar(cls,filename_short,strip=0):
		cls.status_message("Extracting a TAR archive: "+cls.abs_from_short(filename_short))
		filename_full=cls.abs_from_short(filename_short)
		dirname_full=os.path.dirname(filename_full)
		cls.shell('cd "{dir}" && tar --strip-component="{strip}" -xf "{fn}"'.format(dir=dirname_full,strip=strip,fn=filename_full))
		return dirname_full

	@classmethod
	def extract_gz(cls,filename_short):
		cls.status_message("Extracting a GZ archive: "+cls.abs_from_short(filename_short))
		filename_full=cls.abs_from_short(filename_short)
		dirname_full=os.path.dirname(filename_full)
		cls.shell('cd "{dir}" && gzip -d "{fn}"'.format(dir=dirname_full,fn=filename_full))
		return dirname_full

	@classmethod
	def extract_zip(cls,filename_short):
		cls.status_message("Extracting a ZIP archive: "+cls.abs_from_short(filename_short))
		filename_full=cls.abs_from_short(filename_short)
		dirname_full=os.path.dirname(filename_full)
		cls.shell('cd "{dir}" && unzip "{fn}"'.format(dir=dirname_full,fn=filename_full))
		return dirname_full

	@classmethod
	def run_make(cls,dirname_short,clean=False):
		cls.status_message("Running make: "+cls.abs_from_short(dirname_short))
		try:
			cls._run_make(dirname_short=dirname_short,clean=clean,threads=16)
		except:
			try:
				cls._run_make(dirname_short=dirname_short,clean=False,threads=4)
			except:
				try:
					cls._run_make(dirname_short=dirname_short,clean=False,threads=2)
				except:
					cls._run_make(dirname_short=dirname_short,clean=False,threads=1)

	@classmethod
	def _run_make(cls,dirname_short,threads,clean=False):
		dirname_full=cls.abs_from_short(dirname_short)
		other_args=""
		if clean:
			cls.shell('cd "{build_dir}" && make clean'.format(build_dir=dirname_full))
		other_args+=" --jobs {}".format(threads)
		cls.shell('cd "{build_dir}" && make {other_args}'.format(build_dir=dirname_full,other_args=other_args))

	@classmethod
	def run_cmake(cls,dirname_short):
		assert smbl.prog.CMake in cls.depends_on(), "CMake is called but missing in dependencies of the plugin"
		cls.status_message("Running cmake: "+cls.abs_from_short(dirname_short))
		dirname_full=cls.abs_from_short(dirname_short)
		cls.shell('cd "{build_dir}" && "{cmake}" .'.format(
				cmake=smbl.prog.CMAKE,
				build_dir=dirname_full,
			))

	@classmethod
	def run_configure(cls,dirname_short,other_args=""):
		cls.status_message("Running configure: "+cls.abs_from_short(dirname_short))
		dirname_full=cls.abs_from_short(dirname_short)
		cls.shell('cd "{build_dir}" && ./configure {other_args}'.format(build_dir=dirname_full,other_args=other_args))

	@classmethod
	def abs_from_short(cls,short):
		return os.path.abspath(os.path.join(cls.src_dir,short))



##########################################
##########################################


#class Rule:
#	def __init__(self,input,output,run):
#		self.__input=input
#		self.__output=output
#		self.__run=run
#
#		register_rule(self)
#
#	def get_input(self):
#		return self.__input
#
#	def get_output(self):
#		return self.__output
#
#	def run(self):
#		self.__run()
#
#	def encode(self):
#		return "{} {}".format(str(self.__input),str(self.__output))



#
# TODO:
#  - fix sysctl error in CygWin
#


import smbl
import snakemake
import os

from ._program import *

MRSFAST = get_bin_file_path("mrsfast")


##########################################
##########################################


class MrsFast(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				MRSFAST,
			]

	@classmethod
	def install(cls):
		cls.git_clone("http://github.com/sfu-compbio/mrsfast/","")
		cls.run_make("")
		cls.install_file("mrsfast",MRSFAST)

	@classmethod
	def supported_platforms(cls):
		return ["osx","linux"]



import os
import builtins

import smbl.messages

# mapping wrappers
#  - commands in practise:
#       - https://github.com/lh3/mem-paper/blob/master/eval/time.txt
#       - http://lh3lh3.users.sourceforge.net/alnROC.shtml


import sys

try:
	import pkg_resources
	__version__=pkg_resources.get_distribution("smbl").version
except:
	__version__=""

DEFAULT_SMBL_CONF = {
		# print information about SMBL during import
		'print_info': sys.argv[0]=="snakemake",
		# directory where SMBL installs all data and programs
		'directory': os.path.join(os.path.expanduser("~"),".smbl"),
		# create the directory when SMBL is imported
		'ensure_directory': True,
	}

try:
	SMBL_CONF = builtins.SMBL_CONF
except:
	SMBL_CONF = {}

assert type(SMBL_CONF) is dict, "builtins.SMBL_CONF must be a dictionary"
for key in DEFAULT_SMBL_CONF.keys():
	if not key in SMBL_CONF:
		SMBL_CONF[key]=DEFAULT_SMBL_CONF[key]

#print("SMBL configuration:",SMBL_CONF)

if SMBL_CONF["print_info"]:
	import platform
	import sys

	smbl.messages.message("",program="SMBL")
	smbl.messages.message("SnakeMake Bioinformatics Library",program="SMBL")
	smbl.messages.message("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~",program="SMBL")
	smbl.messages.message("Version: {}".format(__version__),program="SMBL")
	smbl.messages.message("Web:     http://github.com/karel-brinda/smbl",program="SMBL")
	smbl.messages.message("Contact: Karel Brinda, karel.brinda@univ-mlv.fr",program="SMBL")
	smbl.messages.message("",program="SMBL")
	smbl.messages.message("Platform: {}".format(sys.platform),program="SMBL")
	smbl.messages.message("System: {}".format(platform.system()),program="SMBL")
	smbl.messages.message("Machine: {}".format(platform.machine()),program="SMBL")
	smbl.messages.message("Processor: {}".format(platform.processor()),program="SMBL")
	smbl.messages.message("Python version: {}".format(platform.python_version()),program="SMBL")
	smbl.messages.message("Python build: {}".format(", ".join(platform.python_build())),program="SMBL")
	smbl.messages.message("",program="SMBL")

smbl_dir = SMBL_CONF["directory"]
bin_dir  = os.path.join(smbl_dir,"bin")
fa_dir   = os.path.join(smbl_dir,"fa")
src_dir  = os.path.join(smbl_dir,"src")

import smbl.utils
import smbl.prog
import smbl.fasta

def include():
	return os.path.join(
			os.path.dirname(__file__),
			"include_all.snake"
		)

def all_fastas():
	return [
			fastafile.get() for fastafile in smbl.fasta.get_registered_fastas()
		]

def all_programs():
	return [
			plugin.get_installation_files() for plugin in smbl.prog.plugins.get_registered_plugins()
		]

def all_compatible_programs():
	return [
			plugin.get_installation_files() for plugin in smbl.prog.plugins.get_registered_plugins()
				if plugin.is_platform_supported()
		]

if SMBL_CONF["ensure_directory"]:
	smbl.utils.shell(
			"""
				mkdir -p "{}" "{}" "{}"
			""".format(bin_dir,fa_dir,src_dir)
		)



import smbl
import snakemake
import os

from ._program import *

ART_454      = get_bin_file_path("art_454")
ART_ILLUMINA = get_bin_file_path("art_illumina")
ART_SOLID    = get_bin_file_path("art_solid")


##########################################
##########################################


class Art(Program):
	@classmethod
	def get_installation_files(cls):
		return [
			ART_454,
			ART_ILLUMINA,
			ART_SOLID,
		]

	@classmethod
	def install(cls):
		if smbl.utils.is_linux()  and smbl.utils.is_os_64bit():
			fn=cls.download_file("http://www.niehs.nih.gov/research/resources/assets/docs/artbinchocolatecherrycake031915linux64tgz.tgz","art.tgz")
		elif smbl.utils.is_linux():
			fn=cls.download_file("http://www.niehs.nih.gov/research/resources/assets/docs/artbinchocolatecherrycake031915linux32tgz.tgz","art.tgz")
		elif smbl.utils.is_osx() and smbl.utils.is_os_64bit():
			fn=cls.download_file("http://www.niehs.nih.gov/research/resources/assets/docs/artbinchocolatecherrycake031915macos64tgz.tgz","art.tgz")
		elif smbl.utils.is_osx():
			fn=cls.download_file("http://www.niehs.nih.gov/research/resources/assets/docs/artbinchocolatecherrycake031915macos32tgz.tgz","art.tgz")
		else:
			smbl.messages.error("Operating system '{}' is not supported".format(smbl.utils.get_platform()),program="SMBL")
			raise NotImplementedError("Unsupported OS")

		dir=cls.extract_tar(fn,strip=2)

		cls.install_file("art_454",ART_454)
		cls.install_file("art_illumina",ART_ILLUMINA)
		cls.install_file("art_SOLiD",ART_SOLID)

	@classmethod
	def supported_platforms(cls):
		return ["osx","linux"]



import smbl
import snakemake
import os

from ._program import *

WGSIM      = get_bin_file_path("wgsim")
WGSIM_EVAL = get_bin_file_path("wgsim_eval.pl")


##########################################
##########################################


class WgSim(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				WGSIM,
				WGSIM_EVAL,
			]

	@classmethod
	def install(cls):
		gitdir=cls.git_clone("http://github.com/lh3/wgsim","wgsim")
		smbl.utils.shell('cd "{dir}" && gcc -g -O2 -Wall -o wgsim wgsim.c -lz -lm'.format(dir=gitdir))
		cls.install_file("wgsim/wgsim",WGSIM)
		cls.install_file("wgsim/wgsim_eval.pl",WGSIM_EVAL)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]



import os
import smbl

from .fastafile import *

EXAMPLE_1  = fasta_file("Mycobacterium_tuberculosis","ftp://ftp.ncbi.nih.gov/genomes/Bacteria/Mycobacterium_tuberculosis_H37Rv_uid170532/NC_018143.fna")
EXAMPLE_2  = fasta_file("Chlamydia_trachomatis","ftp://ftp.ncbi.nih.gov/genomes/Bacteria/Chlamydia_trachomatis_RC_L2_s_3_uid213390/NC_021897.fna")
EXAMPLE_3  = fasta_file("Borrelia_garinii","ftp://ftp.ncbi.nih.gov/genomes/Bacteria/Borrelia_garinii_BgVir_uid162165/NC_017717.fna")

EXAMPLE_42 = fasta_file("Human_chr_21","http://hgdownload.cse.ucsc.edu/goldenpath/hg19/chromosomes/chr21.fa.gz")

EXAMPLE = EXAMPLE_1

HUMAN_HG38 = fasta_file("hg38","http://hgdownload.cse.ucsc.edu/goldenPath/hg38/bigZips/hg38.2bit")
MOUSE_MM10 = fasta_file("mm10","http://hgdownload.cse.ucsc.edu/goldenPath/mm10/bigZips/mm10.2bit")
CHIMP_PANTRO4 = fasta_file("panTro4","http://hgdownload.cse.ucsc.edu/goldenPath/panTro4/bigZips/panTro4.2bit")

for fasta in get_registered_fastas():
	smbl.utils.Rule(
		input=fasta.get_required_programs(),
		output=fasta.get(),
		run=fasta.install_all_steps,
	)



import smbl
import snakemake
import os

from ._program import *

TWOBITTOFA = get_bin_file_path("twoBitToFa")


##########################################
##########################################


class TwoBitToFa(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				TWOBITTOFA
			]

	@classmethod
	def install(cls):
		if smbl.utils.is_linux() and smbl.utils.is_os_64bit():
			url="http://hgdownload.cse.ucsc.edu/admin/exe/linux.x86_64/twoBitToFa"
		elif smbl.utils.is_osx() and smbl.utils.is_os_64bit():
			url="http://hgdownload.cse.ucsc.edu/admin/exe/macOSX.x86_64/twoBitToFa"
		elif smbl.utils.is_osx():
			url="http://hgdownload.cse.ucsc.edu/admin/exe/macOSX.i386/twoBitToFa"
		else:
			smbl.messages.error("Operating system '{}' is not supported".format(smbl.utils.get_platform()),program="SMBL")
			raise NotImplementedError("Unsupported OS")

		cls.download_file(url,"twoBitToFa")
		cls.install_file("twoBitToFa",TWOBITTOFA)

	@classmethod
	def supported_platforms(cls):
		return ["osx","linux"]



#
# TODO:
#  - fix linking error in CygWin
#

import smbl
import snakemake
import os

from ._program import *

from .cmake import *

KALLISTO  = get_bin_file_path("kallisto")


##########################################
##########################################


class Kallisto(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				KALLISTO,
			]

	@classmethod
	def depends_on(cls):
		return [
			CMake,
		]


	@classmethod
	def install(cls):
		gitdir=cls.git_clone("http://github.com/pachterlab/kallisto","kallisto")
		cls.run_cmake("kallisto")
		cls.run_make("kallisto")
		cls.install_file("kallisto/src/kallisto",KALLISTO)

	@classmethod
	def supported_platforms(cls):
		return []
		#["osx","linux","cygwin"]



import smbl
import snakemake
import os

from ._program import *

LASTAL = get_bin_file_path("lastal")
LASTDB = get_bin_file_path("lastdb")


##########################################
##########################################


class Last(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				LASTAL,
				LASTDB,
			]

	@classmethod
	def install(cls):
		last_version="last-638"
		fn=cls.download_file("http://last.cbrc.jp/{}.zip".format(last_version),"last.zip")
		dir1=os.path.dirname(fn)
		smbl.utils.shell('(cd "{dir1}" && unzip last.zip)'.format(dir1=dir1))
		dir2=os.path.join(dir1,last_version)
		cls.run_make(dir2)
		cls.install_file("{}/src/lastal".format(dir2),LASTAL)
		cls.install_file("{}/src/lastdb".format(dir2),LASTDB)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]



#
# TODO:
#  - fix linking error in CygWin
#

import smbl
import snakemake
import os

from ._program import *

DWGSIM      = get_bin_file_path("dwgsim")
DWGSIM_EVAL = get_bin_file_path("dwgsim_eval")


##########################################
##########################################


class DwgSim(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				DWGSIM,
				DWGSIM_EVAL,
			]

	@classmethod
	def install(cls):
		gitdir=cls.git_clone("http://github.com/nh13/dwgsim","dwgsim")
		smbl.prog.correct_samtools_make(os.path.join(gitdir,"samtools","Makefile"))
		cls.run_make("dwgsim")
		cls.install_file("dwgsim/dwgsim",DWGSIM)
		cls.install_file("dwgsim/dwgsim_eval",DWGSIM_EVAL)

	@classmethod
	def supported_platforms(cls):
		return ["osx","linux"]



import smbl
import snakemake
import os

from ._program import *

PERM = get_bin_file_path("perm")


##########################################
##########################################


class PerM(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				PERM,
			]

	@classmethod
	def install(cls):
		if (smbl.utils.is_linux()):
			fn=cls.download_file("https://perm.googlecode.com/files/PerM4_Linux64.gz","perm.gz")
			cls.extract_gz(fn)
			cls.install_file("perm",PERM)
		elif (smbl.utils.is_osx()):
			fn=cls.download_file("https://perm.googlecode.com/files/PerM4_Mac64.tar.gz","perm.tar.gz")
			cls.extract_tar(fn,strip=0)
			cls.install_file("perm",PERM)
		elif (smbl.utils.is_cygwin()):
			fn=cls.download_file("https://perm.googlecode.com/files/PerM4_Win32.zip","perm.zip")
			cls.extract_zip(fn)
			cls.install_file("perm.exe",PERM)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]



import smbl
import snakemake
import os

from ._program import *

TABIX = get_bin_file_path("tabix")
BGZIP = get_bin_file_path("bgzip")


##########################################
##########################################


class HtsLib(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				TABIX,
				BGZIP,
			]

	@classmethod
	def install(cls):
		gitdir_bcftools=cls.git_clone("http://github.com/samtools/bcftools","bcftools")
		gitdir_htslib=cls.git_clone("http://github.com/samtools/htslib","htslib")
		cls.run_make("htslib")
		cls.install_file("htslib/tabix",TABIX)
		cls.install_file("htslib/bgzip",BGZIP)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]



import smbl
import snakemake
import os

from ._program import *

CURESIM = get_bin_file_path("CuReSim.jar")
CURESIM_EVAL = get_bin_file_path("CuReSimEval.jar")


##########################################
##########################################


class CuReSim(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				CURESIM,
				CURESIM_EVAL,
			]

	@classmethod
	def install(cls):
		try:
			fn1=cls.download_file("http://www.pegase-biosciences.com/wp-content/uploads/2013/04/CuReSim1.2.zip","curesim.zip")
			fn2=cls.download_file("http://www.pegase-biosciences.com/wp-content/uploads/2013/04/CuReSimEval1.1.zip","curesim_eval.zip")
			dir=os.path.dirname(fn1)
			smbl.utils.shell('(cd "{dir}" && unzip -j -o curesim.zip && unzip -j -o curesim_eval.zip) > /dev/null'.format(dir=dir))
			cls.install_file("CuReSim.jar",CURESIM)
			cls.install_file("CuReSimEval.jar",CURESIM_EVAL)
		except:
			fn1=cls.download_file("http://github.com/karel-brinda/ext-smbl/raw/master/CuReSim.jar","CuReSim.jar")
			fn2=cls.download_file("http://github.com/karel-brinda/ext-smbl/raw/master/CuReSimEval.jar","CuReSimEval.jar")

		cls.install_file("CuReSim.jar",CURESIM)
		cls.install_file("CuReSimEval.jar",CURESIM_EVAL)


	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]



import smbl
import snakemake
import os

from ._program import *

from .cmake import *

STORM_NUCLEOTIDE = get_bin_file_path("storm-nucleotide")
STORM_COLOR      = get_bin_file_path("storm-color")


##########################################
##########################################


class Storm(Program):

	@classmethod
	def get_installation_files(cls):
		return [
				STORM_NUCLEOTIDE,
				STORM_COLOR,
			]

	@classmethod
	def depends_on(cls):
		return [
			CMake
		]

	@classmethod
	def install(cls):
		cls.svn_checkout("svn://scm.gforge.inria.fr/svnroot/storm/trunk","")
		if smbl.utils.is_osx():
			cls.run_cmake("")
		cls.run_make("")
		cls.install_file("storm-color",STORM_COLOR)
		cls.install_file("storm-nucleotide",STORM_NUCLEOTIDE)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]

	##########################################

	def __init__(
				self,
				fasta,
				bam,
				fastq_1,
				fastq_2=None,
			):

		super().__init__()

		if fastq_2!=None:
			smbl.messages.error("Pair-end reads are not supported by SToRM",program="SMBL",subprogram="SToRM")
			raise NotImplementedError("SToRM supports only single-end reads.")

		self._fa_fn=fasta
		self._fq1_fn=fastq_1
		self._fq2_fn=fastq_2
		self._bam_fn=bam

		smbl.utils.Rule(
			input=self.map_reads_input(),
			output=self.map_reads_output(),
			run=self.map_reads,
		)

	def fq_fn(self):
		if self._fq2_fn==None:
			return [self._fq1_fn]
		else:
			return [self._fq1_fn,self._fq2_fn]

	def fa_fn(self):
		return self._fa_fn

	def bam_fn(self):
		return self._bam_fn

	##########################################

	def map_reads(self,number_of_threads=1):
		
		smbl.utils.shell('"{storm}" -M 4 -A -g "{genome}" -r "{reads}" -N "{threads}" | "{samtools}" view -bS - > "{bam}"'.format(
				storm=STORM_NUCLEOTIDE,
				samtools=smbl.prog.SAMTOOLS,
				genome=self._fa_fn,
				reads=self._fq1_fn,
				bam=self._bam_fn,
				threads=number_of_threads,
			)
		)

	def map_reads_input(self):
		return [
				STORM_NUCLEOTIDE,
				smbl.prog.SAMTOOLS,
				self._fa_fn,
				self._fq1_fn,
			]

	def map_reads_output(self):
		return [
				self.bam_fn(),
			]


import smbl
import snakemake
import os

from ._program import *

BWA = os.path.join(smbl.bin_dir,"bwa")


##########################################
##########################################


class Bwa(Program):

	@classmethod
	def get_installation_files(cls):
		return [BWA]

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]

	@classmethod
	def install(cls):
		cls.git_clone("http://github.com/lh3/bwa","bwa")
		cls.run_make("bwa")
		cls.install_file("bwa/bwa",BWA)

	##########################################

	def __init__(
				self,
				fasta,
				bam,
				fastq_1,
				fastq_2=None,
			):

		super().__init__()


		self._fa_fn=fasta
		self._fq1_fn=fastq_1
		self._fq2_fn=fastq_2
		self._bam_fn=bam

		smbl.utils.Rule(
			input=self.make_index_input(),
			output=self.make_index_output(),
			run=self.make_index,
		)

		smbl.utils.Rule(
			input=self.map_reads_input(),
			output=self.map_reads_output(),
			run=self.map_reads,
		)

	def fq_fn(self):
		if self._fq2_fn==None:
			return [self._fq1_fn]
		else:
			return [self._fq1_fn,self._fq2_fn]

	def fa_fn(self):
		return self._fa_fn

	def bam_fn(self):
		return self._bam_fn

	def index_fns(self):
		return [
			self._fa_fn+".amb",
			self._fa_fn+".ann",
			self._fa_fn+".bwt",
			self._fa_fn+".pac",
			self._fa_fn+".sa"
		]

	##########################################

	def make_index(self):
		smbl.utils.shell('"{bwa}" index {fa}'.format(
				bwa=BWA,
				fa=self._fa_fn,
			))

	def make_index_input(self):
		return [
				BWA,
				self._fa_fn,
			]

	def make_index_output(self):
		return [
				self.index_fns(),
			]

	##########################################

	def map_reads(self):
		smbl.messages.error("Subclass of class Bwa should be used",program="SMBL",subprogram="BWA")
		raise NotImplementedError("Subclass of class Bwa should be used")

	def map_reads_input(self):
		return [
				BWA,
				smbl.prog.SAMTOOLS,
				self.index_fns(),
				self._fa_fn,
				self.fq_fn(),
			]

	def map_reads_output(self):
		return [
				self.bam_fn(),
			]


##########################################
##########################################


class BwaMem(Bwa):
	def __init__(
				self,
				fasta,
				bam,
				fastq_1,
				fastq_2=None,
			):

		super().__init__(
					fasta=fasta,
					fastq_1=fastq_1,
					fastq_2=fastq_2,
					bam=bam,
				)

	def map_reads(self,number_of_threads=1):
		if self._fq2_fn==None:
			reads_string='"{}"'.format(self._fq1_fn)
		else:
			reads_string='"{}" "{}"'.format(self._fq1_fn,self._fq2_fn)

		smbl.utils.shell('"{bwa}" mem -t {threads} "{idx}" {reads_string} | "{samtools}" view -bS - > "{bam}"'.format(
				bwa=BWA,
				samtools=smbl.prog.SAMTOOLS,
				idx=self._fa_fn,
				reads_string=reads_string,
				bam=self._bam_fn,
				threads=number_of_threads,
			)
		)


##########################################
##########################################


class BwaSw(Bwa):
	def __init__(
				self,
				fasta,
				bam,
				fastq_1,
				fastq_2=None,
			):

		super().__init__(
					fasta=fasta,
					fastq_1=fastq_1,
					fastq_2=fastq_2,
					bam=bam,
				)

	def map_reads(self,number_of_threads=1):
		if self._fq2_fn==None:
			reads_string='"{}"'.format(self._fq1_fn)
		else:
			reads_string='"{}" "{}"'.format(self._fq1_fn,self._fq2_fn)

		smbl.utils.shell('"{bwa}" bwasw -t {threads} "{idx}" {reads_string} | "{samtools}" view -bS - > "{bam}"'.format(
				bwa=BWA,
				samtools=smbl.prog.SAMTOOLS,
				idx=self._fa_fn,
				reads_string=reads_string,
				bam=self._bam_fn,
				threads=number_of_threads,
			)
		)



import smbl
import snakemake
import os

from ._program import *

PBSIM = get_bin_file_path("pbsim")


##########################################
##########################################


class PbSim(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				PBSIM,
			]

	@classmethod
	def install(cls):
		fn=cls.download_file("https://pbsim.googlecode.com/files/pbsim-1.0.3.tar.gz","pbsim.tar.gz")
		dir=cls.extract_tar(fn,strip=1)
		cls.run_configure(dir)
		cls.run_make(dir)
		cls.install_file("src/pbsim",PBSIM)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]



import smbl
import snakemake
import os

from ._program import *

BOWTIE2            = get_bin_file_path("bowtie2")
BOWTIE2_ALIGN_L    = get_bin_file_path("bowtie2-align-l")
BOWTIE2_ALIGN_S    = get_bin_file_path("bowtie2-align-s")
BOWTIE2_BUILD      = get_bin_file_path("bowtie2-build")
BOWTIE2_BUILD_L    = get_bin_file_path("bowtie2-build-l")
BOWTIE2_BUILD_S    = get_bin_file_path("bowtie2-build-s")
BOWTIE2_INSPECT    = get_bin_file_path("bowtie2-inspect")
BOWTIE2_INSPECT_L  = get_bin_file_path("bowtie2-inspect-l")
BOWTIE2_INSPECT_S  = get_bin_file_path("bowtie2-inspect-s")


##########################################
##########################################


class Bowtie2(Program):

	@classmethod
	def get_installation_files(cls):
		return [
				BOWTIE2,
				BOWTIE2_ALIGN_L,
				BOWTIE2_ALIGN_S,
				BOWTIE2_BUILD,
				BOWTIE2_BUILD_L,
				BOWTIE2_BUILD_S,
				BOWTIE2_INSPECT,
				BOWTIE2_INSPECT_L,
				BOWTIE2_INSPECT_S,
			]

	@classmethod
	def install(cls):
		gitdir_bcftools=cls.git_clone("http://github.com/BenLangmead/bowtie2","")
		cls.run_make("")
		cls.install_file("bowtie2",BOWTIE2)
		cls.install_file("bowtie2-align-l",BOWTIE2_ALIGN_L)
		cls.install_file("bowtie2-align-s",BOWTIE2_ALIGN_S)
		cls.install_file("bowtie2-build",BOWTIE2_BUILD)
		cls.install_file("bowtie2-build-l",BOWTIE2_BUILD_L)
		cls.install_file("bowtie2-build-s",BOWTIE2_BUILD_S)
		cls.install_file("bowtie2-inspect",BOWTIE2_INSPECT)
		cls.install_file("bowtie2-inspect-l",BOWTIE2_INSPECT_L)
		cls.install_file("bowtie2-inspect-s",BOWTIE2_INSPECT_S)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]

	##########################################

	def __init__(
				self,
				fasta,
				bam,
				fastq_1,
				fastq_2=None,
			):

		super().__init__()


		self._fa_fn=fasta
		self._fq1_fn=fastq_1
		self._fq2_fn=fastq_2
		self._bam_fn=bam

		smbl.utils.Rule(
			input=self.make_index_input(),
			output=self.make_index_output(),
			run=self.make_index,
		)

		smbl.utils.Rule(
			input=self.map_reads_input(),
			output=self.map_reads_output(),
			run=self.map_reads,
		)

	def fq_fn(self):
		if self._fq2_fn==None:
			return [self._fq1_fn]
		else:
			return [self._fq1_fn,self._fq2_fn]

	def fa_fn(self):
		return self._fa_fn

	def bam_fn(self):
		return self._bam_fn

	def index_fns(self):
		return [
			self._fa_fn+".1.bt2",
			self._fa_fn+".2.bt2",
			self._fa_fn+".3.bt2",
			self._fa_fn+".4.bt2",
			self._fa_fn+".rev.1.bt2",
			self._fa_fn+".rev.2.bt2",
		]

	##########################################

	def make_index(self):
		smbl.utils.shell('"{bt2b}" "{fa}" "{fa}"'.format(
				bt2b=BOWTIE2_BUILD,
				fa=self._fa_fn,
			))

	def make_index_input(self):
		return [
				BOWTIE2_BUILD,
				self._fa_fn,
			]

	def make_index_output(self):
		return [
				self.index_fns(),
			]

	##########################################

	def map_reads(self,number_of_threads=1):
		if self._fq2_fn==None:
			reads_string='"{}"'.format(self._fq1_fn)
		else:
			reads_string='-1 "{}" -2 "{}"'.format(self._fq1_fn,self._fq2_fn)

		smbl.utils.shell('"{bt2}" -p {threads} -x "{idx}" {reads_string} | "{samtools}" view -bS - > "{bam}"'.format(
				bt2=BOWTIE2,
				samtools=smbl.prog.SAMTOOLS,
				idx=self._fa_fn,
				reads_string=reads_string,
				bam=self._bam_fn,
				threads=number_of_threads,
			)
		)

	def map_reads_input(self):
		return [
				BOWTIE2,
				smbl.prog.SAMTOOLS,
				self.index_fns(),
				self._fa_fn,
				self.fq_fn(),
			]

	def map_reads_output(self):
		return [
				self.bam_fn(),
			]



import snakemake
import re


def shell(
			cmd,
			remove_spaces=True,
			async=False,
			iterable=False,
			read=False, 
		):
	if remove_spaces:
		#print("removing spaces from command")
		cmd=re.sub(r'[ \t\f\v]+',' ',cmd).strip()

	return snakemake.shell(
			cmd=cmd,
			async=async,
			iterable=iterable,
			read=read,
		)






import smbl
import snakemake
import os

from ._program import *

BFAST = get_bin_file_path("bfast")

#todo: mapping commands from http://genome.jouy.inra.fr/ngs/mapping/


##########################################
##########################################


class BFast(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				BFAST,
			]

	@classmethod
	def install(cls):
		gitdir_bcftools=cls.git_clone("http://github.com/nh13/bfast","")
		smbl.utils.shell('(cd "{}" && sh autogen.sh) > /dev/null'.format(cls.src_dir))
		cls.run_configure("")
		cls.run_make("")
		cls.install_file("bfast/bfast",BFAST)

	@classmethod
	def supported_platforms(cls):
		return ["linux"]



import smbl
import snakemake
import os

import __program

GMAP_BUILD      = __program.get_bin_file_path("gmap_build")
GMAP            = __program.get_bin_file_path("gmapl")
GSNAP           = __program.get_bin_file_path("gsnapl")


##########################################
##########################################


class Gmap_Gsnap(__program.Program):

	version="2014-12-24"

	@classmethod
	def get_installation_files(cls):
		return [
				GMAP,
				GSNAP,
			]

	@classmethod
	def install(cls):
		fn=cls.download_file("http://research-pub.gene.com/gmap/src/gmap-gsnap-{ver}.tar.gz".format(ver=cls.version),"gmap-gsnap.tar.gz")
		dir=cls.extract_tar(fn,strip=1)
		cls.run_configure(dir)
		cls.run_make(dir)
#		cls.install_file("src/gmapindex",GMAP_INDEX)
		cls.install_file("src/gmapl",GMAP)
		cls.install_file("src/gsnapl",GSNAP)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]

	##########################################

	def __init__(
				self,
				fasta,
				bam,
				fastq_1,
				fastq_2=None,
			):

		super().__init__()


		self._fa_fn=fasta
		self._fq1_fn=fastq_1
		self._fq2_fn=fastq_2
		self._bam_fn=bam

#		smbl.prog.plugins.Rule(
#			input=self.make_index_input(),
#			output=self.make_index_output(),
#			run=self.make_index,
#		)

		smbl.prog.plugins.Rule(
			input=self.map_reads_input(),
			output=self.map_reads_output(),
			run=self.map_reads,
		)

	def fq_fn(self):
		if self._fq2_fn==None:
			return [self._fq1_fn]
		else:
			return [self._fq1_fn,self._fq2_fn]

	def fa_fn(self):
		return self._fa_fn

	def bam_fn(self):
		return self._bam_fn

#	def index_fns(self):
#		return [
#			self._fa_fn+".1.bt2",
#			self._fa_fn+".2.bt2",
#			self._fa_fn+".3.bt2",
#			self._fa_fn+".4.bt2",
#			self._fa_fn+".rev.1.bt2",
#			self._fa_fn+".rev.2.bt2",
#		]

	##########################################

	def map_reads(self,number_of_threads=1):
		if self._fq2_fn==None:
			reads_string='"{}"'.format(self._fq1_fn)
		else:
			reads_string='"{}" "{}"'.format(self._fq1_fn,self._fq2_fn)

		snakemake.shell('"{gsnap}" -A sam -d {idx} -t {threads} {reads_string} | "{samtools}" view -bS - > "{bam}"'.format(
				gsnap=GSNAP,
				samtools=smbl.prog.SAMTOOLS,
				idx=self._fa_fn,
				reads_string=reads_string,
				bam=self._bam_fn,
				threads=number_of_threads,
			)
		)

	def map_reads_input(self):
		return [
				GSNAP,
				smbl.prog.SAMTOOLS,
				#self.index_fns(),
				self._fa_fn,
				self.fq_fn(),
			]

	def map_reads_output(self):
		return [
				self.bam_fn(),
			]



import smbl
import snakemake
import os

from ._program import *

SAMBAMBA = get_bin_file_path("sambamba")


##########################################
##########################################


class SamBamBa(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				SAMBAMBA,
			]

	@classmethod
	def install(cls):
		version="v0.5.8"

		if smbl.utils.is_linux():
			fn=cls.download_file("http://github.com/lomereiter/sambamba/releases/download/{ver}/sambamba_{ver}_linux.tar.bz2".format(ver=version),"sambamba.tar.bz2")
		elif smbl.utils.is_osx():
			fn=cls.download_file("http://github.com/lomereiter/sambamba/releases/download/{ver}/sambamba_{ver}_osx.tar.bz2".format(ver=version),"sambamba.tar.bz2")
		else:
			smbl.messages.error("Operating system '{}' is not supported".format(smbl.utils.get_platform()),program="SMBL")
			raise NotImplementedError("Unsupported OS")
		dir=cls.extract_tar(fn)

		cls.install_file("sambamba_{ver}".format(ver=version),SAMBAMBA)

	@classmethod
	def supported_platforms(cls):
		return ["osx","linux"]



import smbl
import snakemake
import os

from ._program import *

SAMTOOLS = get_bin_file_path("samtools")


##########################################
##########################################


class SamTools(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				SAMTOOLS,
			]

	@classmethod
	def install(cls):
		gitdir_samtools=cls.git_clone("http://github.com/samtools/samtools","samtools")
		gitdir_htslib=cls.git_clone("http://github.com/samtools/htslib","htslib")
		smbl.prog.correct_samtools_make(os.path.join(gitdir_samtools,"Makefile"))
		cls.run_make("samtools")
		cls.install_file("samtools/samtools",SAMTOOLS)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]



import smbl
import snakemake
import os

from ._program import *

PICARD = get_bin_file_path("picard.jar")


##########################################
##########################################


class Picard(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				PICARD,
			]

	@classmethod
	def install(cls):
		ver="1.140"
		fn=cls.download_file("https://github.com/broadinstitute/picard/releases/download/{ver}/picard-tools-{ver}.zip".format(ver=ver),"picard.zip")
		dir=os.path.dirname(fn)
		smbl.utils.shell('(cd "{dir}" && unzip -j picard.zip) > /dev/null'.format(dir=dir))
		cls.install_file("picard.jar",PICARD)

	@classmethod
	def supported_platforms(cls):
		return ["osx","linux","cygwin"]



import smbl
import snakemake
import os

from ._program import *

from .cmake import *
from .samtools import *

MASON_FRAG_SEQUENCING   = get_bin_file_path("mason_frag_sequencing")
MASON_GENOME            = get_bin_file_path("mason_genome")
MASON_MATERIALIZER      = get_bin_file_path("mason_materializer")
MASON_METHYLATION       = get_bin_file_path("mason_methylation")
MASON_SIMULATOR         = get_bin_file_path("mason_simulator")
MASON_SPLICING          = get_bin_file_path("mason_splicing")
MASON_VARIATOR          = get_bin_file_path("mason_variator")
YARA_INDEXER            = get_bin_file_path("yara_indexer")
YARA_MAPPER             = get_bin_file_path("yara_mapper")
#RAZERS1                 = get_bin_file_path("razers")
#RAZERS2                 = get_bin_file_path("razers2")
RAZERS3 = RAZER         = get_bin_file_path("razers3")


##########################################
##########################################


class Seqan(Program):

	@classmethod
	def get_installation_files(cls):
		return [
				MASON_FRAG_SEQUENCING,
				MASON_GENOME,
				MASON_MATERIALIZER,
				MASON_METHYLATION,
				MASON_SIMULATOR,
				MASON_SPLICING,
				MASON_VARIATOR,
				YARA_INDEXER,
				YARA_MAPPER,
#				RAZERS1,
#				RAZERS2,
				RAZERS3,
			]

	@classmethod
	def depends_on(cls):
		return [
			CMake,
			SamTools,
		]
		
	@classmethod
	def install(cls):

		cls.git_clone("http://github.com/seqan/seqan","")

		cls.run_cmake("")


		cls.run_make("apps/mason2")
		cls.install_file("bin/mason_frag_sequencing",MASON_FRAG_SEQUENCING)
		cls.install_file("bin/mason_genome",MASON_GENOME)
		cls.install_file("bin/mason_materializer",MASON_MATERIALIZER)
		cls.install_file("bin/mason_methylation",MASON_METHYLATION)
		cls.install_file("bin/mason_simulator",MASON_SIMULATOR)
		cls.install_file("bin/mason_splicing",MASON_SPLICING)
		cls.install_file("bin/mason_variator",MASON_VARIATOR)

#		cls.run_make("apps/razers")
#		cls.install_file("bin/razers",RAZERS1)
#
#		cls.run_make("apps/razers2")
#		cls.install_file("bin/razers2",RAZERS2)

		cls.run_make("apps/razers3")
		cls.install_file("bin/razers3",RAZERS3)

		cls.run_make("apps/yara")
		cls.install_file("bin/yara_indexer",YARA_INDEXER)
		cls.install_file("bin/yara_mapper",YARA_MAPPER)

	@classmethod
	def supported_platforms(cls):
		# cywgin has too old cmake
		return ["osx","linux"]


##########################################
##########################################


class RazerS3(Seqan):

	def __init__(
				self,
				fasta,
				bam,
				fastq_1,
				fastq_2=None,
			):

		super().__init__()


		self._fa_fn=fasta
		self._fq1_fn=fastq_1
		self._fq2_fn=fastq_2
		self._bam_fn=bam

		smbl.utils.Rule(
			input=self.map_reads_input(),
			output=self.map_reads_output(),
			run=self.map_reads,
		)

	def fq_fn(self):
		if self._fq2_fn==None:
			return [self._fq1_fn]
		else:
			return [self._fq1_fn,self._fq2_fn]

	def fa_fn(self):
		return self._fa_fn

	def bam_fn(self):
		return self._bam_fn

	##########################################

	def map_reads(self,number_of_threads=1):
		if self._fq2_fn==None:
			reads_string='"{}"'.format(self._fq1_fn)
		else:
			reads_string='"{}" "{}"'.format(self._fq1_fn,self._fq2_fn)

		smbl.utils.shell('("{razer}" -o "{bam}" -tc {threads} "{genome}" {reads_string}) > /dev/null'.format(
				razer=RAZERS3,
				genome=self._fa_fn,
				reads_string=reads_string,
				bam=self._bam_fn,
				threads=number_of_threads,
			)
		)

	def map_reads_input(self):
		return [
				RAZERS3,
				self._fa_fn,
				self.fq_fn(),
			]

	def map_reads_output(self):
		return [
				self.bam_fn(),
			]


##########################################
##########################################


class Yara(Seqan):

	def __init__(
				self,
				fasta,
				bam,
				fastq_1,
				fastq_2=None,
				sort_by_name=False,
			):

		super().__init__()

		self._fa_fn=fasta
		self._fq1_fn=fastq_1
		self._fq2_fn=fastq_2
		self._bam_fn=bam
		self.index_prefix=self._fa_fn.rpartition(".")[0]
		self._sort_by_name=sort_by_name

		smbl.utils.Rule(
			input=self.make_index_input(),
			output=self.make_index_output(),
			run=self.make_index,
		)

		smbl.utils.Rule(
			input=self.map_reads_input(),
			output=self.map_reads_output(),
			run=self.map_reads,
		)

	def fq_fn(self):
		if self._fq2_fn==None:
			return [self._fq1_fn]
		else:
			return [self._fq1_fn,self._fq2_fn]

	def fa_fn(self):
		return self._fa_fn

	def bam_fn(self):
		return self._bam_fn

	def index_fns(self):
		return [
			self.index_prefix+".lf.drp",
			self.index_prefix+".lf.drs",
			self.index_prefix+".lf.drv",
			self.index_prefix+".lf.pst",
			self.index_prefix+".rid.concat",
			self.index_prefix+".rid.limits",
			self.index_prefix+".sa.ind",
			self.index_prefix+".sa.len",
			self.index_prefix+".sa.val",
			self.index_prefix+".txt.concat",
			self.index_prefix+".txt.limits",
			self.index_prefix+".txt.size",
		]

	##########################################

	def make_index(self):
		smbl.utils.shell('("{yara_indexer}" -o "{prefix}" "{fa}") > /dev/null'.format(
				yara_indexer=YARA_INDEXER,
				prefix=self.index_prefix,
				fa=self._fa_fn,
			))

	def make_index_input(self):
		return [
				YARA_INDEXER,
				self._fa_fn,
			]

	def make_index_output(self):
		return [
				self.index_fns(),
			]

	##########################################

	def map_reads(self,number_of_threads=1):
		if self._fq2_fn==None:
			reads_string='"{}"'.format(self._fq1_fn)
		else:
			reads_string='"{}" "{}"'.format(self._fq1_fn,self._fq2_fn)

		if self._sort_by_name:
			smbl.utils.shell('"{yara_mapper}" -t {threads} "{genome_pref}" {reads_string} | "{samtools}" sort -n - "{bamprefix}"'.format(
						yara_mapper=YARA_MAPPER,
						genome_pref=self.index_prefix,
						reads_string=reads_string,
						bamprefix=self._bam_fn[:-4],
						threads=number_of_threads,
						samtools=smbl.prog.SAMTOOLS,
					)
				)
		else:
			smbl.utils.shell('"{yara_mapper}" -o "{bam}" -t {threads} "{genome_pref}" {reads_string}'.format(
						yara_mapper=YARA_MAPPER,
						genome_pref=self.index_prefix,
						reads_string=reads_string,
						bam=self._bam_fn,
						threads=number_of_threads,
					)
				)

	def map_reads_input(self):
		return [
				YARA_MAPPER,
				self._fa_fn,
				self.fq_fn(),
				self.index_fns(),
			]

	def map_reads_output(self):
		return [
				self.bam_fn(),
			]



#
# TODO:
#  - fix CygWin error
#

import smbl
import snakemake
import os

from ._program import *

GNUPLOT4 = get_bin_file_path("gnuplot4")
GNUPLOT5 = get_bin_file_path("gnuplot5")


##########################################
##########################################


class GnuPlot4(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				GNUPLOT4,
			]

	@classmethod
	def install(cls):
		fn=cls.download_file("http://sourceforge.net/projects/gnuplot/files/gnuplot/4.6.6/gnuplot-4.6.6.tar.gz/download","gnuplot.tar.gz")
		dir=cls.extract_tar(fn,strip=1)
		cls.run_configure(dir,other_args="--with-pdf")
		cls.run_make(dir)
		cls.install_file("src/gnuplot",GNUPLOT4)

	@classmethod
	def supported_platforms(cls):
		return ["osx","linux"]


class GnuPlot5(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				GNUPLOT5,
			]

	@classmethod
	def install(cls):
		fn=cls.download_file("http://sourceforge.net/projects/gnuplot/files/gnuplot/5.0.0/gnuplot-5.0.0.tar.gz/download","gnuplot.tar.gz")
		dir=cls.extract_tar(fn,strip=1)
		cls.run_configure(dir,other_args="--with-pdf")
		cls.run_make(dir)
		cls.install_file("src/gnuplot",GNUPLOT5)

	@classmethod
	def supported_platforms(cls):
		return ["osx","linux"]



import smbl
import snakemake
import os

from ._program import *

SIRFAST = get_bin_file_path("sirfast")


##########################################
##########################################


class SirFast(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				SIRFAST,
			]

	@classmethod
	def install(cls):
		cls.git_clone("http://github.com/BilkentCompGen/sirfast","")
		cls.run_make("")
		cls.install_file("sirfast",SIRFAST)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]



import pytest
import smbl

# todo: ensure that cmake is installed

smbl.prog.CMake.install_all_steps()

@pytest.mark.parametrize("plugin,plugin_name", 
		[
			(x,x.get_plugin_name())
			for x in smbl.prog.plugins.get_registered_plugins() if x.is_platform_supported() and x is not smbl.prog.CMake
		]
	)
def test_plugins(plugin,plugin_name):
	plugin.install_all_steps()
	assert 1==1



import pkgutil
import inspect

from ._program import *

from .art import *
from .bcftools import *
from .bfast import *
from .bowtie import *
from .bwa import *
from .cmake import *
from .curesim import *
from .deez import *
from .drfast import *
from .dwgsim import *
from .freec import *
from .gem import *
from .gnuplot import *
from .htslib import *
from .kallisto import *
from .last import *
from .mrfast import *
from .mrsfast import *
from .pbsim import *
from .perm import *
from .picard import *
from .sambamba import *
from .samtools import *
from .seqan import *
from .sirfast import *
from .storm import *
from .twobittofa import *
from .wgsim import *
from .xs import *



#for loader, name, is_pkg in pkgutil.walk_packages(__path__):
#	module = loader.find_module(name).load_module(name)
#
#	for name, value in inspect.getmembers(module):
#		if name.startswith('__'):
#			continue
#
#		globals()[name] = value
#		__all__.append(name)




import smbl
import snakemake
import os

from ._program import *

XS = get_bin_file_path("XS")


##########################################
##########################################


class Xs(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				XS,
			]

	@classmethod
	def install(cls):
		try:
			fn=cls.download_file("http://bioinformatics.ua.pt/wp-content/uploads/2014/02/XS.tar.gz","xs.tar.gz")
		except:
			fn=cls.download_file("http://exon.ieeta.pt/xs/xs.tar.gz","xs.tar.gz")
		dir=cls.extract_tar(fn,strip=1)
		cls.run_make(dir)
		cls.install_file("XS",XS)

	@classmethod
	def supported_platforms(cls):
		return ["cygwin","osx","linux"]



import smbl
import snakemake
import os
import shutil

__FASTAFILES = set()

def register_fasta(fastafile):
	__FASTAFILES.add(fastafile)

def get_registered_fastas():
	return list(__FASTAFILES)

def get_fa_file_path(fasta):
	return os.path.join(smbl.fa_dir,fasta)

def fasta_file(name,address):
	return FastaFile(name=name,address=address).get()


class FastaFile:
	verbosity=False

	def __init__(self,name,address):
		self.name=name
		self.filename_short=get_fa_file_path("{}.fa".format(name))
		self.filename_full=os.path.abspath(self.filename_short)
		self.address=address
		self.src_dir=os.path.join(smbl.src_dir,self.name)

		if self.address[-3:]==".gz":
			self.format="gz"
		elif self.address[-5:]==".2bit":
			self.format="2bit"
		else:
			self.format="fa"

		register_fasta(self)
		
	def get(self):
		return self.filename_full

	def get_required_programs(self):
		if self.format=="2bit":
			return [smbl.prog.TWOBITTOFA]
		else:
			return []

	def install_all_steps(self):
		self.status_message("Installation started")
		self.install_pre()
		self.install()
		self.install_post()
		self.status_message("Installation completed")

	def install_pre(self):
		try:
			shutil.rmtree(self.src_dir)
		except FileNotFoundError:
			pass
		self.shell('mkdir -p "{src_dir}"'.format(src_dir=self.src_dir))

	def install(self):
		if self.format=="gz":
			archive=self.download(
				filename_full=self.abs_from_short("{}.gz".format(self.name))
			)
			self.extract_gz(
				filename_full_archive=archive,
				filename_full_file=self.filename_full,
			)
		elif self.format=="2bit":
			archive=self.download(
				filename_full=self.abs_from_short("{}.2bit".format(self.name))
			)
			self.extract_2bit(
				filename_full_archive=archive,
				filename_full_file=self.filename_full,
			)
		elif self.format=="fa":
			self.download(
				filename_full=self.filename_full
			)
		else:
			em="Unknown format '{}'".format(self.format)
			smbl.messages.error(em,program="SMBL")
			raise NotImplementedError(em)

	def install_post(self):
		self.shell('rm -fR "{src_dir}"'.format(src_dir=self.src_dir))

	def download(self,filename_full):
		self.status_message("Downloading a file: "+self.address)
		self.shell('mkdir -p "{dir}"'.format(dir=os.path.dirname(filename_full)))
		try:
			self.shell('curl -L --insecure -o "{fn}" "{address}"'.format(fn=filename_full,address=self.address))
		except:
			try:
				self.shell('curl -L --insecure -o "{fn}" "{address}"'.format(fn=filename_full,address=self.address))
			except:
				try:
					self.shell('wget --no-check-certificate --output-document="{fn}" "{address}"'.format(fn=filename_full,address=self.address))
				except:
					self.shell('wget --no-check-certificate --output-document="{fn}" "{address}"'.format(fn=filename_full,address=self.address))
		return filename_full

	def extract_gz(self,filename_full_archive,filename_full_file):
		self.status_message("Extracting an archive: {}".format(filename_full_archive))
		self.shell('gzip -d "{}" -c > "{}"'.format(
				filename_full_archive,
				filename_full_file
			))

	def extract_2bit(self,filename_full_archive,filename_full_file):
		self.status_message("Extracting a 2bit file: {}".format(filename_full_archive))
		self.shell('"{}" "{}" "{}"'.format(
				smbl.prog.TWOBITTOFA,
				filename_full_archive,
				filename_full_file
			))

	def install_file(self,filename_short,dest):
		self.status_message("Moving: "+self.abs_from_short(filename_short))
		filename_full=self.abs_from_short(filename_short)
		self.shell('mv "{source}" "{dest}"'.format(source=filename_full,dest=dest))
		if executable:
			self.shell('chmod +x "{}"'.format(dest))

	def shell(self,command):
		if self.verbosity:
			smbl.utils.shell(command)
		else:
			smbl.utils.shell("({}) > /dev/null".format(command))

	def status_message(self,message):
		smbl.messages.message(
			message=message,
			program="SMBL",
			subprogram=self.name,
		)

	def abs_from_short(self,short):
		return os.path.abspath(os.path.join(self.src_dir,short))

	@classmethod
	def set_verbosity(cls,verbosity):
		cls.verbosity=verbosity		




import sys
import platform

def is_linux():
	return sys.platform.startswith('linux')

def is_cygwin():
	return sys.platform.startswith('cygwin')

def is_windows():
	return sys.platform.startswith('win')

def is_osx():
	return sys.platform.startswith('darwin')

def is_os_64bit():
	return platform.machine().endswith('64')

def get_platform():
	if is_linux():
		return "linux"
	if is_windows():
		return "windows"
	if is_osx():
		return "osx"
	if is_cygwin():
		return "cygwin"



import smbl
import snakemake
import os

from ._program import *

GEM_INDEXER = get_bin_file_path("gem-indexer")
GEM_MAPPER  = get_bin_file_path("gem-mapper")
GEM_2_SAM   = get_bin_file_path("gem-2-sam")


##########################################
##########################################


class Gem(Program):
	@classmethod
	def get_installation_files(cls):
		return [
			GEM_INDEXER,
			GEM_MAPPER,
			GEM_2_SAM,
		]

	@classmethod
	def install(cls):
		if smbl.utils.is_linux() and smbl.utils.is_os_64bit():
			fn=cls.download_file("http://sourceforge.net/projects/gemlibrary/files/gem-library/Binary%20pre-release%203/GEM-binaries-Linux-x86_64-core_i3-20130406-045632.tbz2/download","gem.tbz2")
		else:
			smbl.messages.error("Operating system '{}' is not supported".format(smbl.get_platform()),program="SMBL")
			raise NotImplementedError("Unsupported OS")


		dir=cls.extract_tar(fn,strip=2)


		cls.install_file("gem-indexer",GEM_INDEXER)
		cls.install_file("gem-mapper",GEM_MAPPER)
		cls.install_file("gem-2-sam",GEM_2_SAM)

	@classmethod
	def supported_platforms(cls):
		return ["linux"]



import os
import smbl
#import smbl.prog.plugins

all_platforms = ["linux","windows","osx","cygwin"]

from .plugins import *

#
# create rules from plugins
#
for plugin in get_registered_plugins():
	smbl.utils.Rule(
		input=[
				p.get_installation_files() for p in plugin.depends_on()
			],
		output=plugin.get_installation_files(),
		run=plugin.install_all_steps,
	)



#from smbl.prog.plugins import *
#for i, plugin in enumerate(get_registered_plugins()):
#	rule:
#		output:
#			plugin.get_installation_files()
#		input:
#			[
#				p.get_installation_files() for p in plugin.depends_on()
#			]
#		params:
#			i=str(i),
#		run:
#			(smbl.prog.plugins.get_registered_plugins()[int(params.i)].install_all_steps)()


import smbl

__RULES=set()

def register_rule(rule):
	registered_rules=[r.encode() for r in get_registered_rules()]

	if rule.encode() not in registered_rules:
		__RULES.add(rule)

def get_registered_rules():
	return list(__RULES)

class Rule:
	def __init__(self,input,output,run):
		self.__input=input
		self.__output=output
		self.__run=run

		register_rule(self)

	def get_input(self):
		return self.__input

	def get_output(self):
		return self.__output

	def run(self):
		self.__run()

	def encode(self):
		return "{} {}".format(str(self.__input),str(self.__output))



from termcolor import colored, cprint

def message(message,program=None,subprogram=None):

	if program==None:
		program_part=""
		subprogram_part=""
	else:
		program_part="[{}] ".format(program)
		if subprogram==None:
			subprogram_part=""
		else:
			subprogram_part="{}: ".format(subprogram)

	cprint(
		"".join([program_part,subprogram_part,message]),
		"blue",
		attrs=['bold'],
	)

def error(message,program=None,subprogram=None, exception=None):
	if exception!=None:
		assert issubclass(exception,Exception)

	if program==None:
		program_part=""
		subprogram_part=""
	else:
		program_part="[{}] ".format(program)
		if subprogram==None:
			subprogram_part=""
		else:
			subprogram_part="{}: ".format(subprogram)

	cprint(
		"".join([program_part,subprogram_part,"Error: ",message]),
		"red",
		attrs=['bold'],
	)

	if exception!=None:
		raise exception(message)



import smbl
import snakemake
import os

from ._program import *

FREEC = get_bin_file_path("freec")


##########################################
##########################################


class Freec(Program):
	@classmethod
	def get_installation_files(cls):
		return [
				FREEC,
			]

	@classmethod
	def install(cls):
		fn=cls.download_file("http://bioinfo-out.curie.fr/projects/freec/src/FREEC_Linux64.tar.gz","freec.tar.gz")
		dir=cls.extract_tar(fn)
		cls.run_make(dir,clean=True)
		cls.install_file("freec",FREEC)

	@classmethod
	def supported_platforms(cls):
		return ["linux"]
